# Test cases for bedrock_server_manager.config.settings
import pytest
import os
import json
import shutil
import logging  # Import logging
from pathlib import Path

# Assuming the module is in src/bedrock_server_manager/config/settings.py
# Adjust the import path if your project structure is different.
# This might require adding src to PYTHONPATH or using relative imports if tests are part of the package.
from bedrock_server_manager.config.settings import (
    Settings,
    deep_merge,
    CONFIG_SCHEMA_VERSION,
    NEW_CONFIG_FILE_NAME,
    OLD_CONFIG_FILE_NAME,
)
from bedrock_server_manager.config.const import package_name, env_name
from bedrock_server_manager.error import ConfigurationError


# Helper function to create a dummy config file
def create_dummy_config(config_path: Path, content: dict):
    config_path.parent.mkdir(parents=True, exist_ok=True)
    with open(config_path, "w") as f:
        json.dump(content, f, indent=4)


# Helper function to read a config file
def read_json_file(file_path: Path) -> dict:
    with open(file_path, "r") as f:
        return json.load(f)


@pytest.fixture
def tmp_config_dir(tmp_path: Path) -> Path:
    """Creates a temporary directory for config files."""
    # tmp_path is a built-in pytest fixture providing a Path object to a temporary directory
    return tmp_path


@pytest.fixture
def settings_instance(tmp_config_dir: Path, monkeypatch) -> Settings:
    """
    Creates a Settings instance using a temporary config directory.
    Temporarily overrides environment variables for BSM_DATA_DIR and cleans up.
    """
    # Define a temporary data directory within the test's tmp_path structure
    test_data_dir = tmp_config_dir / "test_app_data"
    test_data_dir.mkdir(parents=True, exist_ok=True)

    # Temporarily set the BSM_DATA_DIR environment variable using the correct env_name
    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(test_data_dir))

    # The Settings class constructor will use this env var to determine paths
    settings = Settings()

    # Ensure the settings object is using the correct temporary path
    # This is more of an assertion for the fixture's correctness
    assert settings.app_data_dir == str(test_data_dir)
    assert settings.config_dir == str(test_data_dir / ".config")
    assert settings.config_path == str(test_data_dir / ".config" / NEW_CONFIG_FILE_NAME)

    yield settings

    # Teardown: monkeypatch automatically undoes env var changes.
    # shutil.rmtree(test_data_dir) # tmp_path handles cleanup of its contents


def test_settings_initialization_default(
    settings_instance: Settings, tmp_config_dir: Path
):
    """Test Settings initialization with default values when no config file exists."""
    config_file = Path(settings_instance.config_path)

    assert config_file.exists()
    loaded_settings = read_json_file(config_file)

    assert loaded_settings["config_version"] == CONFIG_SCHEMA_VERSION
    assert loaded_settings["paths"]["servers"] == str(
        tmp_config_dir / "test_app_data" / "servers"
    )
    assert loaded_settings["retention"]["backups"] == 3
    # Check a few default values
    assert settings_instance.get("retention.backups") == 3
    assert settings_instance.get("web.port") == 11325
    assert settings_instance.get("non_existent.key") is None
    assert settings_instance.get("non_existent.key", "default_val") == "default_val"


def test_settings_load_existing_config(
    settings_instance: Settings, tmp_config_dir: Path
):
    """Test Settings loading from an existing v2 config file."""
    custom_config_content = {
        "config_version": CONFIG_SCHEMA_VERSION,
        "paths": {
            "servers": str(tmp_config_dir / "custom_server_path"),  # Use temp path
            "logs": str(tmp_config_dir / "custom_logs"),
        },
        "retention": {"backups": 10},
        "logging": {"cli_level": 10},  # Example: logging.DEBUG
        "custom": {"my_setting": "my_value"},
    }
    config_file = Path(settings_instance.config_path)
    create_dummy_config(config_file, custom_config_content)

    settings_instance.reload()  # Reload to pick up the created file

    assert settings_instance.get("paths.servers") == str(
        tmp_config_dir / "custom_server_path"
    )
    assert settings_instance.get("paths.logs") == str(tmp_config_dir / "custom_logs")
    assert settings_instance.get("paths.content") == str(
        Path(settings_instance.app_data_dir) / "content"
    )  # check default value is still there, relative to correct app_data_dir
    assert settings_instance.get("retention.backups") == 10
    assert settings_instance.get("logging.cli_level") == 10
    assert settings_instance.get("custom.my_setting") == "my_value"
    # Ensure directories mentioned in config are created
    assert Path(settings_instance.get("paths.logs")).exists()


def test_settings_set_and_save(settings_instance: Settings):
    """Test setting a value and saving it to the config file."""
    settings_instance.set("web.host", "0.0.0.0")
    settings_instance.set("retention.logs", 7)
    settings_instance.set("new_category.new_setting.nested", "nested_value")

    config_file = Path(settings_instance.config_path)
    loaded_settings = read_json_file(config_file)

    assert loaded_settings["web"]["host"] == "0.0.0.0"
    assert loaded_settings["retention"]["logs"] == 7
    assert loaded_settings["new_category"]["new_setting"]["nested"] == "nested_value"

    # Test that setting the same value doesn't rewrite (not easily testable without mocks for _write_config)
    # For now, just ensure the value remains correct
    settings_instance.set("web.host", "0.0.0.0")
    assert settings_instance.get("web.host") == "0.0.0.0"


def test_settings_migrate_old_filename(tmp_config_dir: Path, monkeypatch):
    """Test migration from old config filename to new config filename."""
    test_data_dir = tmp_config_dir / "test_app_data_migration"
    test_data_dir.mkdir(parents=True, exist_ok=True)
    config_dir_path = test_data_dir / ".config"
    config_dir_path.mkdir(parents=True, exist_ok=True)

    old_config_file = config_dir_path / OLD_CONFIG_FILE_NAME
    new_config_file = config_dir_path / NEW_CONFIG_FILE_NAME

    # Create a dummy old config file (can be v1 or v2 content, filename is key)
    dummy_content = {"config_version": CONFIG_SCHEMA_VERSION, "web": {"port": 12345}}
    create_dummy_config(old_config_file, dummy_content)

    assert old_config_file.exists()
    assert not new_config_file.exists()

    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(test_data_dir))
    settings = Settings()  # Instantiation should trigger migration

    assert not old_config_file.exists()  # Old file should be renamed
    assert new_config_file.exists()
    assert settings.config_path == str(new_config_file)
    assert settings.get("web.port") == 12345


def test_settings_migrate_v1_to_v2(tmp_config_dir: Path, monkeypatch):
    """Test migration from v1 (flat) config to v2 (nested) config."""
    test_data_dir = tmp_config_dir / "test_app_data_v1_migration"
    test_data_dir.mkdir(parents=True, exist_ok=True)
    config_dir_path = test_data_dir / ".config"
    config_dir_path.mkdir(parents=True, exist_ok=True)

    config_file = config_dir_path / NEW_CONFIG_FILE_NAME

    # Create a dummy v1 config file
    v1_servers_path = (
        test_data_dir / "v1_servers"
    )  # Use a path within the temp structure
    v1_content = {"BASE_DIR": str(v1_servers_path), "BACKUP_KEEP": 5, "WEB_PORT": 8000}
    create_dummy_config(config_file, v1_content)

    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(test_data_dir))
    settings = Settings()  # Instantiation should trigger migration

    backup_v1_file = Path(str(config_file) + ".v1.bak")
    assert backup_v1_file.exists()  # Old v1 file should be backed up

    assert settings.get("config_version") == CONFIG_SCHEMA_VERSION
    assert settings.get("paths.servers") == str(v1_servers_path)
    assert settings.get("retention.backups") == 5
    assert settings.get("web.port") == 8000
    # Check a default v2 value that wasn't in v1
    assert (
        settings.get("logging.cli_level")
        == settings.default_config["logging"]["cli_level"]
    )

    loaded_settings = read_json_file(config_file)
    assert loaded_settings["config_version"] == CONFIG_SCHEMA_VERSION
    assert loaded_settings["paths"]["servers"] == str(v1_servers_path)


def test_settings_ensure_dirs_created(settings_instance: Settings):
    """Test that critical directories are created."""
    assert Path(settings_instance.get("paths.servers")).exists()
    assert Path(settings_instance.get("paths.content")).exists()
    assert Path(settings_instance.get("paths.downloads")).exists()
    assert Path(settings_instance.get("paths.backups")).exists()
    assert Path(settings_instance.get("paths.plugins")).exists()
    assert Path(settings_instance.get("paths.logs")).exists()

    # Test with a custom path not created by default init but by set
    custom_dir_path = Path(settings_instance.app_data_dir) / "custom_test_dir"
    assert not custom_dir_path.exists()
    settings_instance.set("paths.custom_test", str(custom_dir_path))
    # _ensure_dirs_exist is called on load, but if we add a new path type,
    # it's not automatically created by 'set'. This is expected.
    # The _ensure_dirs_exist is more about the core paths.
    # If a path is added to the "paths" section of default_config and then created,
    # it *would* be created by _ensure_dirs_exist upon next load.

    # Let's test if a new path added to the structure and then reloaded gets created
    # This means we need to modify the _settings directly for this test, or save and reload
    new_log_path_val = str(Path(settings_instance.app_data_dir) / ".new_logs")
    settings_instance._settings["paths"]["new_logs_test"] = new_log_path_val
    settings_instance._write_config()  # Write the modified config
    settings_instance.reload()  # Reload, which calls _ensure_dirs_exist

    # This test is a bit tricky because _ensure_dirs_exist only iterates specific keys.
    # A more robust test would involve mocking os.makedirs and checking calls.
    # For now, we rely on the explicit list in _ensure_dirs_exist.
    # The existing test for default paths is sufficient.


def test_settings_reload(settings_instance: Settings):
    """Test reloading settings from the config file."""
    settings_instance.set("web.port", 11111)  # Initial set

    config_file_path = Path(settings_instance.config_path)
    current_config = read_json_file(config_file_path)
    current_config["web"]["port"] = 22222
    current_config["custom"]["reloaded_val"] = True
    create_dummy_config(config_file_path, current_config)  # Simulate external change

    settings_instance.reload()

    assert settings_instance.get("web.port") == 22222
    assert settings_instance.get("custom.reloaded_val") is True


def test_settings_app_data_dir_override(tmp_path: Path, monkeypatch):
    """Test that BSM_DATA_DIR environment variable correctly overrides default app data dir."""
    custom_data_path = tmp_path / "custom_bsm_data"
    # No need to mkdir, Settings should do it.

    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(custom_data_path))
    settings = Settings()

    assert settings.app_data_dir == str(custom_data_path)
    assert settings.config_dir == str(custom_data_path / ".config")
    assert settings.config_path == str(
        custom_data_path / ".config" / NEW_CONFIG_FILE_NAME
    )
    assert custom_data_path.exists()
    assert (custom_data_path / ".config").exists()
    assert Path(settings.config_path).exists()


def test_settings_default_paths_without_env_override(tmp_path: Path, monkeypatch):
    """Test default path determination when BSM_DATA_DIR is not set."""
    # Ensure the env var is not set for this test
    monkeypatch.delenv(f"{package_name.upper()}_DATA_DIR", raising=False)

    # We need to control where the home directory is perceived to be for consistent testing
    # We'll patch os.path.expanduser for this test
    fake_home_dir = tmp_path / "fake_user_home"
    fake_home_dir.mkdir()

    original_expanduser = os.path.expanduser

    def mock_expanduser(path):
        if path == "~":
            return str(fake_home_dir)
        return original_expanduser(path)

    monkeypatch.setattr(os.path, "expanduser", mock_expanduser)

    settings = Settings()

    expected_data_dir = fake_home_dir / package_name
    assert settings.app_data_dir == str(expected_data_dir)
    assert settings.config_dir == str(expected_data_dir / ".config")
    assert (expected_data_dir / ".config" / NEW_CONFIG_FILE_NAME).exists()


def test_deep_merge():
    """Test the deep_merge utility function."""
    dest = {"a": 1, "b": {"c": 2, "d": 3, "x": {"y": 10}}, "f": 7}
    src = {"b": {"c": 5, "e": 6, "x": {"z": 20}}, "g": 8}

    merged = deep_merge(src, dest)

    assert merged["a"] == 1
    assert merged["b"]["c"] == 5  # Overwritten
    assert merged["b"]["d"] == 3  # Preserved
    assert merged["b"]["e"] == 6  # Added
    assert merged["b"]["x"]["y"] == 10  # Preserved nested
    assert merged["b"]["x"]["z"] == 20  # Added nested
    assert merged["f"] == 7
    assert merged["g"] == 8  # Added
    assert dest is merged  # Check modified in place

    # Test merging into a non-dict node
    dest2 = {"a": 1, "b": "not a dict"}
    src2 = {"b": {"c": 2}}
    merged2 = deep_merge(src2, dest2)
    assert merged2["b"]["c"] == 2

    # Test merging with empty source
    dest3 = {"a": 1}
    src3 = {}
    merged3 = deep_merge(src3, dest3)
    assert merged3["a"] == 1

    # Test merging with empty destination
    dest4 = {}
    src4 = {"a": 1}
    merged4 = deep_merge(src4, dest4)
    assert merged4["a"] == 1


def test_settings_load_corrupted_json(settings_instance: Settings, caplog):
    """Test loading a corrupted JSON config file."""
    config_file = Path(settings_instance.config_path)
    with open(config_file, "w") as f:
        f.write("this is not valid json")

    settings_instance.reload()  # Attempt to load corrupted file

    assert "Could not load config file" in caplog.text
    assert "Using default settings" in caplog.text
    # Check that settings are reset to defaults
    assert (
        settings_instance.get("retention.backups")
        == settings_instance.default_config["retention"]["backups"]
    )
    assert (
        settings_instance.get("web.port")
        == settings_instance.default_config["web"]["port"]
    )

    # Verify that a subsequent 'set' operation saves a valid default config
    settings_instance.set("web.threads", 8)
    assert settings_instance.get("web.threads") == 8
    try:
        loaded_settings = read_json_file(config_file)
        assert loaded_settings["web"]["threads"] == 8
        assert (
            loaded_settings["config_version"] == CONFIG_SCHEMA_VERSION
        )  # Should be a full valid config
    except json.JSONDecodeError:
        pytest.fail(
            "Config file was not rewritten with valid JSON after corruption and set operation."
        )


def test_config_file_permission_error_on_write(
    settings_instance: Settings, monkeypatch, caplog, tmp_path: Path
):
    """Test handling of permission errors when writing config file."""

    # Make the config file read-only to simulate permission issue for writing
    config_file_path = Path(settings_instance.config_path)

    # For this to work consistently, especially on POSIX, the directory needs to be writable
    # but the file itself not. If the directory is not writable, os.rename in migrate might fail first.
    # We will mock open to raise OSError for the write operation.

    original_open = open

    def mock_open_raise_oserror(file, mode="r", *args, **kwargs):
        if str(file) == str(config_file_path) and "w" in mode:
            raise OSError("Permission denied for testing")
        return original_open(file, mode, *args, **kwargs)

    monkeypatch.setattr("builtins.open", mock_open_raise_oserror)

    with pytest.raises(
        ConfigurationError,
        match="Failed to write configuration: Permission denied for testing",
    ):
        settings_instance.set("web.port", 9999)  # This should attempt to write

    # Test that an error during initial _write_config in load (e.g. if file doesn't exist) is also handled.
    # For this, we need a fresh Settings instance where the file doesn't exist yet.

    # Create a scenario for a new Settings instance with a distinct app_data_dir
    # Use a sub-directory of tmp_path to ensure it's cleaned up
    app_data_for_new_instance = tmp_path / "new_instance_app_data"
    app_data_for_new_instance.mkdir(parents=True, exist_ok=True)

    # The Settings class will create ".config" inside this new app_data_dir.
    # We need to ensure this ".config" directory exists so that only open() fails.
    config_dir_for_new_instance = app_data_for_new_instance / ".config"
    config_dir_for_new_instance.mkdir(parents=True, exist_ok=True)

    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(app_data_for_new_instance))

    # Determine the config path that the *new* Settings instance will use
    expected_new_config_path = config_dir_for_new_instance / NEW_CONFIG_FILE_NAME

    # The mock_open_raise_oserror_multi is already set from the previous part of the test.
    # It needs to correctly identify this new expected_new_config_path.

    # We need a more targeted mock for the second part.
    # The previous mock_open_raise_oserror is specific to the settings_instance.
    # The mock_open_raise_oserror_multi was an attempt to combine them. Let's ensure it's correctly defined for this new path.
    # or make a new one if it's cleaner. For now, let's make it conditional.

    # Store original config_file_path for the first part of the test
    original_mocked_path = config_file_path

    def mock_open_raise_oserror_multi(file, mode="r", *args, **kwargs):
        # Case 1: For the settings_instance (first part of the test)
        if str(file) == str(original_mocked_path) and "w" in mode:
            raise OSError("Permission denied for testing (settings_instance)")
        # Case 2: For the new Settings() instance (second part of the test)
        elif str(file) == str(expected_new_config_path) and "w" in mode:
            raise OSError("Permission denied for testing (new Settings())")
        return original_open(file, mode, *args, **kwargs)

    monkeypatch.setattr("builtins.open", mock_open_raise_oserror_multi)

    # Re-run the first part of the test with the new multi-mock
    # (This part was already passing, but good to ensure it still does)
    with pytest.raises(
        ConfigurationError,
        match="Failed to write configuration: Permission denied for testing \(settings_instance\)",
    ):
        settings_instance.set(
            "web.port", 9998
        )  # Use a different port to ensure 'set' runs

    # The Settings constructor calls load(), which calls _write_config() if file not found.
    with pytest.raises(
        ConfigurationError,
        match="Failed to write configuration: Permission denied for testing \(new Settings\(\)\)",
    ):
        Settings()


def test_config_file_migration_rename_error(tmp_config_dir: Path, monkeypatch, caplog):
    """Test handling of OSError during config filename migration."""
    caplog.set_level(logging.INFO)
    test_data_dir = tmp_config_dir / "test_migration_rename_fail"
    config_dir_path = test_data_dir / ".config"
    config_dir_path.mkdir(parents=True, exist_ok=True)

    old_config_file = config_dir_path / OLD_CONFIG_FILE_NAME
    new_config_file = config_dir_path / NEW_CONFIG_FILE_NAME  # Should not exist yet

    create_dummy_config(old_config_file, {"some_key": "some_value"})  # Old file exists

    original_rename = os.rename

    def mock_rename_raise_oserror(src, dst):
        if str(src) == str(old_config_file) and str(dst) == str(new_config_file):
            raise OSError("Permission denied for renaming (testing)")
        return original_rename(src, dst)

    monkeypatch.setattr(os, "rename", mock_rename_raise_oserror)
    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(test_data_dir))

    with pytest.raises(
        ConfigurationError, match=r"Failed to rename configuration file"
    ) as excinfo:
        Settings()

    assert "Failed to rename configuration file" in str(excinfo.value)
    assert old_config_file.exists()  # Old file should still be there
    assert (
        not new_config_file.exists()
    )  # New file should not have been created by rename


def test_v1_migration_backup_rename_error(tmp_config_dir: Path, monkeypatch, caplog):
    """Test handling of OSError during v1 config backup before migration."""
    caplog.set_level(logging.INFO)
    test_data_dir = tmp_config_dir / "test_v1_backup_fail"
    config_dir_path = test_data_dir / ".config"
    config_dir_path.mkdir(parents=True, exist_ok=True)

    config_file = (
        config_dir_path / NEW_CONFIG_FILE_NAME
    )  # V1 content, but new name (e.g. after filename migration)
    v1_content = {"BASE_DIR": "/v1/servers"}  # V1 content (no config_version)
    create_dummy_config(config_file, v1_content)

    backup_path = Path(str(config_file) + ".v1.bak")

    original_rename = os.rename

    def mock_rename_raise_oserror_for_backup(src, dst):
        if str(src) == str(config_file) and str(dst) == str(backup_path):
            raise OSError("Permission denied for v1 backup rename (testing)")
        return original_rename(src, dst)

    monkeypatch.setattr(os, "rename", mock_rename_raise_oserror_for_backup)
    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(test_data_dir))

    with pytest.raises(
        ConfigurationError, match=r"Failed to back up old config file"
    ) as excinfo:
        Settings()

    assert "Failed to back up old config file" in str(excinfo.value)
    assert config_file.exists()  # Original v1 file should still be there
    assert not backup_path.exists()  # Backup should not have been created
    # And settings should likely be default, or the process halted.
    # Check that the file content is still v1.
    current_content = read_json_file(config_file)
    assert "config_version" not in current_content
    assert current_content["BASE_DIR"] == "/v1/servers"


def test_ensure_dirs_permission_error(settings_instance: Settings, monkeypatch, caplog):
    """Test handling of OSError when creating critical directories."""
    caplog.set_level(logging.INFO)
    # We need to make one of the target directories unwritable by its parent.
    # This is hard to do without affecting other tests or requiring sudo.
    # A more common scenario is that os.makedirs itself fails.

    original_makedirs = os.makedirs
    failing_path = settings_instance.get("paths.logs")  # Pick one path to fail

    def mock_makedirs_raise_oserror(path, exist_ok=True):
        if str(Path(path)) == str(Path(failing_path)):
            raise OSError(f"Cannot create directory {path} (testing)")
        return original_makedirs(path, exist_ok=exist_ok)

    monkeypatch.setattr(os, "makedirs", mock_makedirs_raise_oserror)

    # To trigger _ensure_dirs_exist, we reload.
    # However, settings_instance already called it in __init__.
    # We need a fresh instance or a way to force _ensure_dirs_exist.
    # Let's modify the path and reload.

    # Create a new settings instance in a new temp dir for this test
    # to ensure _ensure_dirs_exist is called in a controlled way.
    temp_dir_for_ensure_test = settings_instance.app_data_dir + "_ensure_fail"
    Path(temp_dir_for_ensure_test).mkdir(
        parents=True, exist_ok=True
    )  # Create the base app data dir

    # The .config dir inside it should NOT be created yet by this test setup,
    # so Settings() will try to create it, and then the log dir.

    monkeypatch.setenv(f"{env_name}_DATA_DIR", str(temp_dir_for_ensure_test))

    # The failing_path needs to be relative to this new data dir.
    # Re-calculate based on the new settings that will be created.
    expected_new_log_path = str(Path(temp_dir_for_ensure_test) / ".logs")

    # Adjust the mock_makedirs to use the correct path for this new instance
    def mock_makedirs_raise_oserror_specific(path, exist_ok=True):
        if str(Path(path)) == expected_new_log_path:
            raise OSError(f"Cannot create directory {path} (testing)")
        # Allow .config to be created
        elif str(Path(path)) == str(Path(temp_dir_for_ensure_test) / ".config"):
            return original_makedirs(path, exist_ok=exist_ok)
        return original_makedirs(path, exist_ok=exist_ok)

    monkeypatch.setattr(os, "makedirs", mock_makedirs_raise_oserror_specific)

    with pytest.raises(
        ConfigurationError,
        match=rf"Could not create critical directory: {expected_new_log_path}",
    ) as excinfo:
        Settings()  # This will call _ensure_dirs_exist

    assert f"Could not create critical directory: {expected_new_log_path}" in str(
        excinfo.value
    )



# bedrock_server_manager/cli/server_permissions.py
"""
Defines the `bsm permissions` command group for managing player permission levels.

This module provides CLI tools to view and set player permission levels
(e.g., member, operator, visitor) for a specific Bedrock server. These
permissions are typically stored in the server's ``permissions.json`` file.

The commands interact with the API layer, specifically functions in
:mod:`~bedrock_server_manager.api.server_install_config` for getting and
setting permissions, and :mod:`~bedrock_server_manager.api.player` for
looking up player XUIDs from the global player database.

Key functionalities:

    -   An interactive workflow (:func:`~.interactive_permissions_workflow`) to
        guide users through selecting a player and assigning a permission level.
    -   Direct commands (``bsm permissions set``, ``bsm permissions list``) for
        scripting or quick modifications.

"""
from typing import Optional

import click
import questionary

from ..api import player as player_api, server_install_config as config_api
from .utils import (
    handle_api_response as _handle_api_response,
)
from ..error import BSMError


def interactive_permissions_workflow(server_name: str):
    """
    Guides the user through an interactive workflow to set a player's permission level.

    This function performs the following steps:

        1.  Fetches all known players from the global player database using
            :func:`~bedrock_server_manager.api.player.get_all_known_players_api`.
        2.  If no players are found, it informs the user and suggests commands to
            populate the database.
        3.  Prompts the user to select a player from the list using `questionary`.
        4.  Prompts the user to select a permission level (member, operator, visitor)
            for the chosen player.
        5.  Calls :func:`~bedrock_server_manager.api.server_install_config.configure_player_permission`
            to apply the selected permission to the specified server.
        6.  Uses :func:`~.handle_api_response` to display the outcome.

    Args:
        server_name (str): The name of the server for which to configure permissions.

    Raises:
        click.Abort: If the user cancels the operation at any `questionary` prompt
                     (e.g., by pressing Ctrl+C).
    """
    click.secho("\n--- Interactive Permission Configuration ---", bold=True)
    player_response = player_api.get_all_known_players_api()
    all_players = player_response.get("players", [])

    if not all_players:
        click.secho(
            "No players found in the global player database (players.json).",
            fg="yellow",
        )
        click.secho(
            "Run 'bsm player scan' or 'bsm player add' to populate it first.", fg="cyan"
        )
        return

    # Create a user-friendly mapping for the selection prompt
    player_map = {f"{p['name']} (XUID: {p['xuid']})": p for p in all_players}
    choices = sorted(list(player_map.keys())) + ["Cancel"]

    player_choice_str = questionary.select(
        "Select a player to configure permissions for:", choices=choices
    ).ask()

    if not player_choice_str or player_choice_str == "Cancel":
        raise click.Abort()

    selected_player = player_map[player_choice_str]
    permission = questionary.select(
        f"Select permission level for {selected_player['name']}:",
        choices=["member", "operator", "visitor"],
        default="member",
    ).ask()

    if permission is None:  # User pressed Ctrl+C
        raise click.Abort()

    perm_response = config_api.configure_player_permission(
        server_name, selected_player["xuid"], selected_player["name"], permission
    )
    _handle_api_response(
        perm_response,
        f"Permission for {selected_player['name']} set to '{permission}'.",
    )


@click.group()
def permissions():
    """
    Manages player permission levels (e.g., operator, member) on a server.

    These commands interact with the server's `permissions.json` file.
    """
    pass


@permissions.command("set")
@click.option(
    "-s",
    "--server",
    "server_name",
    required=True,
    help="The name of the target server.",
)
@click.option(
    "-p",
    "--player",
    "player_name",
    help="The gamertag of the player. Skips interactive mode.",
)
@click.option(
    "-l",
    "--level",
    type=click.Choice(["visitor", "member", "operator"], case_sensitive=False),
    help="The permission level to grant. Skips interactive mode.",
)
def set_perm(server_name: str, player_name: Optional[str], level: Optional[str]):
    """
    Sets a permission level for a player on a specific server.

    If both `--player` and `--level` options are provided, this command directly
    attempts to set the permission. It first looks up the player's XUID from the
    global player database using their gamertag.

    If either `--player` or `--level` is omitted, the command falls back to an
    interactive workflow (:func:`~.interactive_permissions_workflow`) to guide
    the user through selecting a player and permission level.

    Permission levels are typically 'visitor', 'member', or 'operator'.

    Calls APIs:
        - :func:`~bedrock_server_manager.api.player.get_all_known_players_api` (for XUID lookup)
        - :func:`~bedrock_server_manager.api.server_install_config.configure_player_permission`
    """
    try:
        if not player_name or not level:
            click.secho(
                f"Player or level not specified; starting interactive editor for '{server_name}'...",
                fg="yellow",
            )
            interactive_permissions_workflow(server_name)
            return

        # Direct, non-interactive logic
        click.echo(f"Finding player '{player_name}' in global database...")
        all_players_resp = player_api.get_all_known_players_api()
        player_data = next(
            (
                p
                for p in all_players_resp.get("players", [])
                if p.get("name", "").lower() == player_name.lower()
            ),
            None,
        )

        if not player_data or not player_data.get("xuid"):
            click.secho(
                f"Error: Player '{player_name}' not found in the global player database.",
                fg="red",
            )
            click.secho(
                "Run 'bsm player add' or 'bsm player scan' to add them first.",
                fg="cyan",
            )
            raise click.Abort()

        xuid = player_data["xuid"]
        click.echo(
            f"Setting permission for {player_name} (XUID: {xuid}) to '{level}'..."
        )
        response = config_api.configure_player_permission(
            server_name, xuid, player_name, level
        )
        _handle_api_response(response, "Permission updated successfully.")

    except (click.Abort, KeyboardInterrupt, BSMError):
        click.secho("\nOperation cancelled.", fg="yellow")


@permissions.command("list")
@click.option(
    "-s", "--server", "server_name", required=True, help="The name of the server."
)
def list_perms(server_name: str):
    """
    Lists all configured player permissions for a specific server.

    Retrieves and displays the contents of the server's `permissions.json` file,
    showing each player's gamertag (if known from the global player DB), XUID,
    and their assigned permission level (e.g., Operator, Member).

    Permission levels are color-coded for readability.

    Calls API: :func:`~bedrock_server_manager.api.server_install_config.get_server_permissions_api`.
    """
    response = config_api.get_server_permissions_api(server_name)

    # Handle API errors first
    if response.get("status") == "error":
        _handle_api_response(response, "")
        return

    permissions = response.get("data", []).get("permissions", [])

    if not permissions:
        click.secho(
            f"The permissions file for server '{server_name}' is empty or does not exist.",
            fg="yellow",
        )
        return

    click.secho(f"\nPermissions for '{server_name}':", bold=True)
    for p in permissions:
        # Use styled output for permission levels for better readability
        level = p.get(
            "permission_level", "unknown"
        ).lower()  # API used `permission_level` before, now `permission`
        level_color = {"operator": "red", "member": "green", "visitor": "blue"}.get(
            level, "white"
        )
        level_styled = click.style(level.capitalize(), fg=level_color, bold=True)

        name = p.get("name", "Unknown Player")
        xuid = p.get("xuid", "N/A")
        click.echo(f"  - {name:<20} (XUID: {xuid:<18}) {level_styled}")



# bedrock_server_manager/api/server.py
"""Provides API functions for managing Bedrock server instances.

This module serves as a key interface layer for server-specific operations within
the Bedrock Server Manager. It leverages the
:class:`~bedrock_server_manager.core.bedrock_server.BedrockServer` core class
to perform a variety of actions such as server lifecycle management (starting,
stopping, restarting), configuration (getting/setting server-specific properties),
and command execution.

The functions within this module are designed to return structured dictionary
responses, making them suitable for consumption by web API routes, command-line
interface (CLI) commands, or other parts of the application. This module also
integrates with the plugin system by exposing many of its functions as callable
APIs for plugins (via :func:`~bedrock_server_manager.plugins.api_bridge.plugin_method`)
and by triggering various plugin events during server operations.
"""

import os
import logging
from typing import Dict, Any
import platform
import shutil
import subprocess

# Guarded import for Windows-specific functionality
if platform.system() == "Windows":
    try:
        import win32serviceutil

        PYWIN32_AVAILABLE = True
    except ImportError:
        PYWIN32_AVAILABLE = False
else:
    PYWIN32_AVAILABLE = False


# Plugin system imports to bridge API functionality.
from ..plugins import plugin_method

# Local application imports.
from ..instances import get_server_instance, get_plugin_manager_instance
from ..config import EXPATH
from ..config import API_COMMAND_BLACKLIST
from ..core.system import (
    launch_detached_process,
    get_bedrock_launcher_pid_file_path,
    remove_pid_file_if_exists,
)
from ..error import (
    BSMError,
    InvalidServerNameError,
    UserInputError,
    ServerError,
    BlockedCommandError,
    MissingArgumentError,
)

logger = logging.getLogger(__name__)

plugin_manager = get_plugin_manager_instance()


@plugin_method("get_server_setting")
def get_server_setting(server_name: str, key: str) -> Dict[str, Any]:
    """Reads any value from a server's specific JSON configuration file
    (e.g., ``<server_name>_config.json``) using dot-notation for keys.

    Args:
        server_name (str): The name of the server.
        key (str): The dot-notation key to read from the server's JSON
            configuration (e.g., "server_info.status", "settings.autoupdate",
            "custom.my_value").

    Returns:
        Dict[str, Any]: A dictionary containing the operation result.
        On success: ``{"status": "success", "value": <retrieved_value>}``
        On error: ``{"status": "error", "message": "<error_message>"}``
        The ``<retrieved_value>`` will be ``None`` if the key is not found.

    Raises:
        InvalidServerNameError: If `server_name` is empty.
        MissingArgumentError: If `key` is empty.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")
    if not key:
        raise MissingArgumentError("A 'key' must be provided.")

    logger.debug(f"API: Reading server setting for '{server_name}': Key='{key}'")
    try:
        server = get_server_instance(server_name)
        # Use the internal method to access any key
        value = server._manage_json_config(key, "read")
        return {"status": "success", "value": value}
    except BSMError as e:
        logger.error(
            f"API: Error reading setting '{key}' for server '{server_name}': {e}"
        )
        return {"status": "error", "message": str(e)}
    except Exception as e:
        logger.error(
            f"API: Unexpected error reading setting for '{server_name}': {e}",
            exc_info=True,
        )
        return {"status": "error", "message": "An unexpected error occurred."}


def set_server_setting(server_name: str, key: str, value: Any) -> Dict[str, Any]:
    """Writes any value to a server's specific JSON configuration file
    (e.g., ``<server_name>_config.json``) using dot-notation for keys.
    Intermediate dictionaries will be created if they don't exist along the key path.

    Args:
        server_name (str): The name of the server.
        key (str): The dot-notation key to write to in the server's JSON
            configuration (e.g., "server_info.status", "custom.new_setting").
        value (Any): The new value to write. Must be JSON serializable.

    Returns:
        Dict[str, Any]: A dictionary containing the operation result.
        On success: ``{"status": "success", "message": "Setting '<key>' updated..."}``
        On error: ``{"status": "error", "message": "<error_message>"}``

    Raises:
        InvalidServerNameError: If `server_name` is empty.
        MissingArgumentError: If `key` is empty.
        ConfigParseError: If `value` is not JSON serializable or if an
            intermediate part of the `key` path conflicts with an existing
            non-dictionary item.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")
    if not key:
        raise MissingArgumentError("A 'key' must be provided.")

    logger.info(
        f"API: Writing server setting for '{server_name}': Key='{key}', Value='{value}'"
    )
    try:
        server = get_server_instance(server_name)
        # Use the internal method to write to any key
        server._manage_json_config(key, "write", value)
        return {
            "status": "success",
            "message": f"Setting '{key}' updated for server '{server_name}'.",
        }
    except BSMError as e:
        logger.error(f"API: Error setting '{key}' for server '{server_name}': {e}")
        return {"status": "error", "message": str(e)}
    except Exception as e:
        logger.error(
            f"API: Unexpected error setting value for '{server_name}': {e}",
            exc_info=True,
        )
        return {"status": "error", "message": "An unexpected error occurred."}


@plugin_method("set_server_custom_value")
def set_server_custom_value(server_name: str, key: str, value: Any) -> Dict[str, Any]:
    """Writes a key-value pair to the 'custom' section of a server's specific
    JSON configuration file (e.g., ``<server_name>_config.json``).
    This is a sandboxed way for plugins or users to store arbitrary data
    associated with a server. The key will be stored as ``custom.<key>``.

    Args:
        server_name (str): The name of the server.
        key (str): The key (string) for the custom value within the 'custom' section.
            Cannot be empty.
        value (Any): The value to write. Must be JSON serializable.

    Returns:
        Dict[str, Any]: A dictionary containing the operation result.
        On success: ``{"status": "success", "message": "Custom value '<key>' updated..."}``
        On error: ``{"status": "error", "message": "<error_message>"}``

    Raises:
        InvalidServerNameError: If `server_name` is empty.
        MissingArgumentError: If `key` is empty.
        ConfigParseError: If `value` is not JSON serializable.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")
    if not key:
        raise MissingArgumentError("A 'key' must be provided.")

    logger.info(f"API (Plugin): Writing custom value for '{server_name}': Key='{key}'")
    try:
        server = get_server_instance(server_name)
        # This method is sandboxed to the 'custom' section
        server.set_custom_config_value(key, value)
        return {
            "status": "success",
            "message": f"Custom value '{key}' updated for server '{server_name}'.",
        }
    except BSMError as e:
        logger.error(
            f"API (Plugin): Error setting custom value for '{server_name}': {e}"
        )
        return {"status": "error", "message": str(e)}
    except Exception as e:
        logger.error(
            f"API (Plugin): Unexpected error setting custom value for '{server_name}': {e}",
            exc_info=True,
        )
        return {"status": "error", "message": "An unexpected error occurred."}


@plugin_method("get_all_server_settings")
def get_all_server_settings(server_name: str) -> Dict[str, Any]:
    """Reads the entire JSON configuration for a specific server from its
    dedicated configuration file (e.g., ``<server_name>_config.json``).
    If the file doesn't exist, it will be created with default values.
    Handles schema migration if an older config format is detected.

    Args:
        server_name (str): The name of the server.

    Returns:
        Dict[str, Any]: A dictionary containing the operation result.
        On success: ``{"status": "success", "data": <all_settings_dict>}``
        On error: ``{"status": "error", "message": "<error_message>"}``

    Raises:
        InvalidServerNameError: If `server_name` is empty.
        FileOperationError: If creating/reading the config directory/file fails.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")

    logger.debug(f"API: Reading all settings for server '{server_name}'.")
    try:
        server = get_server_instance(server_name)
        # _load_server_config handles loading and migration
        all_settings = server._load_server_config()
        return {"status": "success", "data": all_settings}
    except BSMError as e:
        logger.error(f"API: Error reading all settings for server '{server_name}': {e}")
        return {"status": "error", "message": str(e)}
    except Exception as e:
        logger.error(
            f"API: Unexpected error reading all settings for '{server_name}': {e}",
            exc_info=True,
        )
        return {"status": "error", "message": "An unexpected error occurred."}


@plugin_method("start_server")
def start_server(
    server_name: str,
    mode: str = "direct",
) -> Dict[str, Any]:
    """Starts the specified Bedrock server.

    Triggers the ``before_server_start`` and ``after_server_start`` plugin events.
    Manages platform-specific start methods:

    -   **direct**: Runs the server directly in the current process via
        :meth:`~.core.bedrock_server.BedrockServer.start`. This is a blocking
        call until the server stops.
    -   **detached**: Attempts to start the server in the background. It prioritizes
        using the OS-native service manager (systemd on Linux, Windows Services on
        Windows) if a service for this server is configured and active. If a
        service is not used or fails, it falls back to launching a new,
        independent background process via
        :func:`~.core.system.process.launch_detached_process`.

    Args:
        server_name (str): The name of the server to start.
        mode (str, optional): The start mode. Can be 'direct' or 'detached'.
            Defaults to 'direct'.

    Returns:
        Dict[str, Any]: A dictionary containing the operation result.
        - If direct mode: ``{"status": "success", "message": "Server... (direct mode) process finished."}``
        - If detached via service: ``{"status": "success", "message": "Server... started via <service_manager>."}``
        - If detached via fallback: ``{"status": "success", "message": "Server... start initiated... (Launcher PID: <pid>).", "pid": <pid>}``
        - On error: ``{"status": "error", "message": "<error_message>"}``

    Raises:
        InvalidServerNameError: If `server_name` is not provided.
        UserInputError: If `mode` is invalid (not 'direct' or 'detached').
        ServerStartError: If the server is not installed, already running, or if
            :meth:`~.core.bedrock_server.BedrockServer.start` (in direct mode)
            encounters an issue.
        BSMError: For other application-specific errors during startup.
    """
    mode = mode.lower()

    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")
    if mode not in ["direct", "detached"]:
        raise UserInputError(
            f"Invalid start mode '{mode}'. Must be 'direct' or 'detached'."
        )

    # --- Plugin Hook ---
    plugin_manager.trigger_guarded_event(
        "before_server_start", server_name=server_name, mode=mode
    )

    logger.info(f"API: Attempting to start server '{server_name}' in '{mode}' mode...")
    result = {}
    try:
        server = get_server_instance(server_name)

        if server.is_running():
            logger.warning(
                f"API: Server '{server_name}' is already running. Start request ignored."
            )
            return {
                "status": "error",
                "message": f"Server '{server_name}' is already running.",
            }

        if mode == "direct":
            logger.debug(
                f"API: Calling server.start() for '{server_name}' (direct mode)."
            )
            server.start()  # This is a blocking call.
            logger.info(f"API: Direct start for server '{server_name}' completed.")
            result = {
                "status": "success",
                "message": f"Server '{server_name}' (direct mode) process finished.",
            }
            return result
        elif mode == "detached":
            use_service_manager = False

            # --- OS-Specific Service Start (Preferred Method) ---
            if platform.system() == "Linux" and server.check_service_exists():
                logger.debug(f"API: Using systemd to start server '{server_name}'.")
                systemctl_cmd_path = shutil.which("systemctl")
                if systemctl_cmd_path:
                    try:
                        subprocess.run(
                            [
                                systemctl_cmd_path,
                                "--user",
                                "start",
                                server.systemd_service_name_full,
                            ],
                            check=True,
                            capture_output=True,
                            text=True,
                        )
                        use_service_manager = True
                        result = {
                            "status": "success",
                            "message": f"Server '{server_name}' started via systemd.",
                        }
                    except subprocess.CalledProcessError as e:
                        logger.warning(
                            f"systemd service '{server.systemd_service_name_full}' failed to start: {e.stderr.strip()}. "
                            "Falling back to generic detached process."
                        )
                else:
                    logger.warning(
                        "'systemctl' command not found, falling back to generic detached process."
                    )
            elif (
                platform.system() == "Windows"
                and PYWIN32_AVAILABLE
                and server.check_service_exists()
            ):
                logger.debug(f"API: Using Windows Service to start '{server_name}'.")
                try:
                    win32serviceutil.StartService(server.windows_service_name)
                    use_service_manager = True
                    result = {
                        "status": "success",
                        "message": f"Server '{server_name}' started via Windows Service.",
                    }
                except Exception as e:
                    logger.warning(
                        f"Windows service '{server.windows_service_name}' failed to start: {e}. "
                        "Falling back to generic detached process."
                    )

            if use_service_manager:
                return result

            # --- Generic Detached Start (Fallback for All OSes) ---
            logger.info(
                f"API: Starting server '{server_name}' using generic detached process launcher."
            )
            cli_command_parts = [
                EXPATH,
                "server",
                "start",
                "--server",
                server_name,
                "--mode",
                "direct",  # The detached process runs the server directly
            ]
            cli_command_str_list = [os.fspath(part) for part in cli_command_parts]
            launcher_pid_file_path = get_bedrock_launcher_pid_file_path(
                server.server_name,
                server.server_config_dir,  # Use BedrockServer instance properties
            )

            launcher_pid = launch_detached_process(
                cli_command_str_list,
                launcher_pid_file_path,  # Pass the launcher PID file path
            )
            logger.info(
                f"API: Detached server starter for '{server_name}' launched with PID {launcher_pid}."
            )
            result = {
                "status": "success",
                "message": f"Server '{server_name}' start initiated in detached mode (Launcher PID: {launcher_pid}).",
                "pid": launcher_pid,
            }
            return result

    except BSMError as e:
        logger.error(f"API: Failed to start server '{server_name}': {e}", exc_info=True)
        result = {
            "status": "error",
            "message": f"Failed to start server '{server_name}': {e}",
        }
        return result
    except Exception as e:
        logger.error(
            f"API: Unexpected error starting server '{server_name}': {e}", exc_info=True
        )
        result = {
            "status": "error",
            "message": f"Unexpected error starting server '{server_name}': {e}",
        }
        return result
    finally:
        # --- Plugin Hook ---
        plugin_manager.trigger_guarded_event(
            "after_server_start", server_name=server_name, result=result
        )


@plugin_method("stop_server")
def stop_server(server_name: str) -> Dict[str, str]:
    """Stops the specified Bedrock server.

    Triggers the ``before_server_stop`` and ``after_server_stop`` plugin events.
    The method prioritizes stopping via the OS-native service manager (systemd on
    Linux, Windows Services on Windows) if the server's service is active.
    If not managed by a service or if service stop fails, it falls back to a
    direct stop attempt using
    :meth:`~.core.bedrock_server.BedrockServer.stop`, which involves sending
    a "stop" command and then potentially forcefully terminating the process.

    Args:
        server_name (str): The name of the server to stop.

    Returns:
        Dict[str, str]: A dictionary containing the operation result.

        On success: ``{"status": "success", "message": "Server... stopped successfully."}`` or
                    ``{"status": "success", "message": "Server... stop initiated via <service_manager>."}``

        On error (e.g., already stopped): ``{"status": "error", "message": "<error_message>"}``

    Raises:
        InvalidServerNameError: If `server_name` is not provided.
        ServerStopError: If the server fails to stop after all attempts.
        BSMError: For other application-specific errors during shutdown.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")

    # --- Plugin Hook ---
    plugin_manager.trigger_guarded_event("before_server_stop", server_name=server_name)

    logger.info(f"API: Attempting to stop server '{server_name}'...")
    result = {}
    try:
        server = get_server_instance(server_name)

        if not server.is_running():
            logger.warning(
                f"API: Server '{server_name}' is not running. Stop request ignored."
            )
            server.set_status_in_config("STOPPED")
            result = {
                "status": "error",
                "message": f"Server '{server_name}' was already stopped.",
            }
            return result

        # --- OS-Specific Service Stop (Preferred Method) ---
        service_stop_initiated = False
        if platform.system() == "Linux" and server.is_service_active():
            logger.debug(f"API: Attempting to stop '{server_name}' using systemd...")
            try:
                systemctl_cmd_path = shutil.which("systemctl")
                subprocess.run(
                    [
                        systemctl_cmd_path,
                        "--user",
                        "stop",
                        server.systemd_service_name_full,
                    ],
                    check=True,
                    capture_output=True,
                    text=True,
                )
                service_stop_initiated = True
                result = {
                    "status": "success",
                    "message": f"Server '{server_name}' stop initiated via systemd.",
                }
            except (subprocess.CalledProcessError, FileNotFoundError) as e:
                logger.warning(
                    f"API: Stopping via systemd failed: {e}. Falling back to direct stop."
                )
        elif platform.system() == "Windows" and server.is_service_active():
            logger.debug(
                f"API: Attempting to stop '{server_name}' via Windows Service..."
            )
            try:
                win32serviceutil.StopService(server.windows_service_name)
                service_stop_initiated = True
                result = {
                    "status": "success",
                    "message": f"Server '{server_name}' stop initiated via Windows Service.",
                }
            except Exception as e:
                logger.warning(
                    f"API: Stopping via Windows Service failed: {e}. Falling back to direct stop."
                )
        server.stop()
        logger.info(f"API: Server '{server_name}' stopped successfully.")
        result = {
            "status": "success",
            "message": f"Server '{server_name}' stopped successfully.",
        }

        try:
            launcher_pid_file = get_bedrock_launcher_pid_file_path(
                server.server_name, server.server_config_dir
            )
            remove_pid_file_if_exists(launcher_pid_file)
        except Exception as e_launcher_cleanup:
            logger.debug(
                f"Error during launcher PID cleanup for '{server_name}': {e_launcher_cleanup}"
            )

        return result

    except BSMError as e:
        logger.error(f"API: Failed to stop server '{server_name}': {e}", exc_info=True)
        result = {
            "status": "error",
            "message": f"Failed to stop server '{server_name}': {e}",
        }
        return result
    except Exception as e:
        logger.error(
            f"API: Unexpected error stopping server '{server_name}': {e}", exc_info=True
        )
        result = {
            "status": "error",
            "message": f"Unexpected error stopping server '{server_name}': {e}",
        }
        return result
    finally:
        # --- Plugin Hook ---
        plugin_manager.trigger_guarded_event(
            "after_server_stop", server_name=server_name, result=result
        )


@plugin_method("restart_server")
def restart_server(server_name: str, send_message: bool = True) -> Dict[str, str]:
    """Restarts the specified Bedrock server by orchestrating stop and start.

    This function internally calls :func:`~.stop_server` and then
    :func:`~.start_server` (with ``mode="detached"``).

    - If the server is already stopped, this function will attempt to start it
      in 'detached' mode.
    - If running, it will attempt to stop it (optionally sending a restart
      message to the server if ``send_message=True``), wait briefly for the
      stop to complete, and then start it again in 'detached' mode.

    Args:
        server_name (str): The name of the server to restart.
        send_message (bool, optional): If ``True``, attempts to send a "say Restarting server..."
            message to the server console via
            :meth:`~.core.bedrock_server.BedrockServer.send_command`
            before stopping. Defaults to ``True``.

    Returns:
        Dict[str, str]: A dictionary with the operation status and a message,
        reflecting the outcome of the start/stop operations.
        On success: ``{"status": "success", "message": "Server... restarted successfully."}``
        On error: ``{"status": "error", "message": "Restart failed: <reason>"}``

    Raises:
        InvalidServerNameError: If `server_name` is not provided.
        ServerStartError: If the start phase fails (from :func:`~.start_server`).
        ServerStopError: If the stop phase fails (from :func:`~.stop_server`).
        BSMError: For other application-specific errors.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")

    logger.debug(
        f"API: Initiating restart for server '{server_name}'. Send message: {send_message}"
    )
    try:
        server = get_server_instance(server_name)
        is_running = server.is_running()

        # If server is not running, just start it.
        if not is_running:
            logger.info(
                f"API: Server '{server_name}' was not running. Attempting to start..."
            )
            start_result = start_server(server_name, mode="detached")
            if start_result.get("status") == "success":
                start_result["message"] = (
                    f"Server '{server_name}' was not running and has been started."
                )
            return start_result

        # If server is running, perform the stop-start cycle.
        logger.info(
            f"API: Server '{server_name}' is running. Proceeding with stop/start cycle."
        )
        if send_message:
            try:
                server.send_command("say Restarting server...")
            except BSMError as e:
                logger.warning(
                    f"API: Failed to send restart warning to '{server_name}': {e}"
                )

        stop_result = stop_server(server_name)
        if stop_result.get("status") == "error":
            stop_result["message"] = (
                f"Restart failed during stop phase: {stop_result.get('message')}"
            )
            return stop_result

        start_result = start_server(server_name, mode="detached")
        if start_result.get("status") == "error":
            start_result["message"] = (
                f"Restart failed during start phase: {start_result.get('message')}"
            )
            return start_result

        logger.info(f"API: Server '{server_name}' restarted successfully.")
        return {
            "status": "success",
            "message": f"Server '{server_name}' restarted successfully.",
        }

    except BSMError as e:
        logger.error(
            f"API: Failed to restart server '{server_name}': {e}", exc_info=True
        )
        return {"status": "error", "message": f"Restart failed: {e}"}
    except Exception as e:
        logger.error(
            f"API: Unexpected error during restart for '{server_name}': {e}",
            exc_info=True,
        )
        return {"status": "error", "message": f"Unexpected error during restart: {e}"}


@plugin_method("send_command")
def send_command(server_name: str, command: str) -> Dict[str, str]:
    """Sends a command to a running Bedrock server.

    The command is checked against a blacklist (defined by
    :const:`~bedrock_server_manager.config.blocked_commands.API_COMMAND_BLACKLIST`)
    before being sent via
    :meth:`~.core.bedrock_server.BedrockServer.send_command`.
    Triggers ``before_command_send`` and ``after_command_send`` plugin events.

    Args:
        server_name (str): The name of the server to send the command to.
        command (str): The command string to send (e.g., "list", "say Hello").
            Cannot be empty.

    Returns:
        Dict[str, str]: On successful command submission, returns a dictionary:
        ``{"status": "success", "message": "Command '<command>' sent successfully."}``.
        If an error occurs, an exception is raised instead of returning an error dictionary.

    Raises:
        InvalidServerNameError: If `server_name` is not provided.
        MissingArgumentError: If `command` is empty.
        BlockedCommandError: If the command is in the API blacklist.
        ServerNotRunningError: If the target server is not running.
        SendCommandError: For underlying issues during command transmission (e.g., pipe errors).
        ServerError: For other unexpected errors during the operation.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")
    if not command or not command.strip():
        raise MissingArgumentError("Command cannot be empty.")

    command_clean = command.strip()

    # --- Plugin Hook ---
    plugin_manager.trigger_event(
        "before_command_send", server_name=server_name, command=command_clean
    )

    logger.info(
        f"API: Attempting to send command to server '{server_name}': '{command_clean}'"
    )
    result = {}
    try:
        # Check command against the configured blacklist.
        blacklist = API_COMMAND_BLACKLIST or []
        command_check = command_clean.lower().lstrip("/")
        for blocked_cmd_prefix in blacklist:
            if isinstance(blocked_cmd_prefix, str) and command_check.startswith(
                blocked_cmd_prefix.lower()
            ):
                error_msg = f"Command '{command_clean}' is blocked by configuration."
                logger.warning(
                    f"API: Blocked command attempt for '{server_name}': {error_msg}"
                )
                raise BlockedCommandError(error_msg)

        server = get_server_instance(server_name)
        server.send_command(command_clean)

        logger.info(
            f"API: Command '{command_clean}' sent successfully to server '{server_name}'."
        )
        result = {
            "status": "success",
            "message": f"Command '{command_clean}' sent successfully.",
        }
        return result

    except BSMError as e:
        logger.error(
            f"API: Failed to send command to server '{server_name}': {e}", exc_info=True
        )
        # Re-raise to allow higher-level handlers to catch specific BSM errors.
        raise
    except Exception as e:
        logger.error(
            f"API: Unexpected error sending command to '{server_name}': {e}",
            exc_info=True,
        )
        # Wrap unexpected errors in a generic ServerError.
        raise ServerError(f"Unexpected error sending command: {e}") from e
    finally:
        # --- Plugin Hook ---
        plugin_manager.trigger_event(
            "after_command_send",
            server_name=server_name,
            command=command_clean,
            result=result,
        )


def delete_server_data(
    server_name: str, stop_if_running: bool = True
) -> Dict[str, str]:
    """Deletes all data associated with a Bedrock server.

    .. danger::
        This is a **HIGHLY DESTRUCTIVE** and irreversible operation.

    It calls :meth:`~.core.bedrock_server.BedrockServer.delete_all_data`, which
    removes:
    - The server's main installation directory.
    - The server's JSON configuration subdirectory.
    - The server's entire backup directory.
    - The server's systemd user service file (Linux) or Windows Service entry.
    - The server's PID file.

    Triggers ``before_delete_server_data`` and ``after_delete_server_data`` plugin events.

    Args:
        server_name (str): The name of the server to delete.
        stop_if_running (bool, optional): If ``True`` (default), the server will be
            stopped using :func:`~.stop_server` before its data is deleted.
            If ``False`` and the server is running, the operation will likely
            fail due to file locks or other conflicts.

    Returns:
        Dict[str, str]: A dictionary with the operation status and a message.
        On success: ``{"status": "success", "message": "All data for server... deleted successfully."}``
        On error: ``{"status": "error", "message": "<error_message>"}``

    Raises:
        InvalidServerNameError: If `server_name` is not provided.
        ServerStopError: If `stop_if_running` is ``True`` and the server fails to stop.
        FileOperationError: If deleting one or more essential directories or files fails.
        BSMError: For other application-specific errors.
    """
    if not server_name:
        raise InvalidServerNameError("Server name cannot be empty.")

    # --- Plugin Hook ---
    plugin_manager.trigger_event("before_delete_server_data", server_name=server_name)

    # High-visibility warning for a destructive operation.
    logger.warning(
        f"API: !!! Initiating deletion of ALL data for server '{server_name}'. Stop if running: {stop_if_running} !!!"
    )
    result = {}
    try:
        server = get_server_instance(server_name)

        # Stop the server first if requested and it's running.
        if stop_if_running and server.is_running():
            logger.info(
                f"API: Server '{server_name}' is running. Stopping before deletion..."
            )

            stop_result = stop_server(server_name)
            if stop_result.get("status") == "error":
                error_msg = f"Failed to stop server '{server_name}' before deletion: {stop_result.get('message')}. Deletion aborted."
                logger.error(error_msg)
                result = {"status": "error", "message": error_msg}
                return result

            logger.info(f"API: Server '{server_name}' stopped.")

        # Attempt to remove the associated system service before deleting files.
        if server.check_service_exists():
            logger.info(f"API: Removing system service for '{server_name}'...")
            try:
                server.disable_service()
                server.remove_service()
                logger.info(f"API: System service for '{server_name}' removed.")
            except BSMError as e:
                logger.warning(
                    f"API: Could not remove system service for '{server_name}': {e}. Continuing with data deletion."
                )

        logger.debug(
            f"API: Proceeding with deletion of data for server '{server_name}'..."
        )
        server.delete_all_data()
        logger.info(f"API: Successfully deleted all data for server '{server_name}'.")
        result = {
            "status": "success",
            "message": f"All data for server '{server_name}' deleted successfully.",
        }
        return result

    except BSMError as e:
        logger.error(
            f"API: Failed to delete server data for '{server_name}': {e}", exc_info=True
        )
        result = {"status": "error", "message": f"Failed to delete server data: {e}"}
        return result
    except Exception as e:
        logger.error(
            f"API: Unexpected error deleting server data for '{server_name}': {e}",
            exc_info=True,
        )
        result = {
            "status": "error",
            "message": f"Unexpected error deleting server data: {e}",
        }
        return result
    finally:
        # --- Plugin Hook ---
        plugin_manager.trigger_event(
            "after_delete_server_data", server_name=server_name, result=result
        )



# bedrock_server_manager/cli/web.py
"""
Defines the `bsm web` command group for managing the Bedrock Server Manager Web UI.

This module provides CLI commands to control the lifecycle of the web server
application (FastAPI/Uvicorn based) and to manage its integration as an
OS-level system service (e.g., systemd on Linux, Windows Services on Windows).

Key command groups and commands include:

    -   ``bsm web start``: Starts the web server, with options for host, port,
        debug mode, and detached/direct execution.
    -   ``bsm web stop``: Stops a detached web server process.
    -   ``bsm web service ...``: A subgroup for managing the Web UI's system service:
        -   ``bsm web service configure``: Interactively or directly configures the
            Web UI system service (creation, autostart).
        -   ``bsm web service enable``: Enables the Web UI service for autostart.
        -   ``bsm web service disable``: Disables autostart for the Web UI service.
        -   ``bsm web service remove``: Removes the Web UI system service definition.
        -   ``bsm web service status``: Checks the status of the Web UI system service.

Interactions with system services are contingent on the availability of
appropriate service management tools on the host OS (e.g., `systemctl` for
systemd, `pywin32` for Windows Services). The commands use functions from
:mod:`~bedrock_server_manager.api.web` and the
:class:`~bedrock_server_manager.core.manager.BedrockServerManager`.
"""

import functools
import logging
from typing import Tuple, Callable, Optional

import click
import questionary

from ..api import web as web_api
from .utils import handle_api_response as _handle_api_response
from ..core import BedrockServerManager
from ..error import (
    BSMError,
    MissingArgumentError,
)

logger = logging.getLogger(__name__)


# --- Web System Service ---
def requires_web_service_manager(func: Callable) -> Callable:
    """
    A decorator to ensure Web UI service management commands only run on capable systems.

    This decorator checks if the system has a supported service manager
    (e.g., systemd for Linux, or `pywin32` installed for Windows services)
    by inspecting `bsm.can_manage_services` from the
    :class:`~.core.manager.BedrockServerManager` instance in the Click context.

    If the capability is not present, it prints an error and aborts the command.

    Args:
        func (Callable): The Click command function to decorate.

    Returns:
        Callable: The wrapped command function.
    """

    @functools.wraps(func)
    @click.pass_context
    def wrapper(ctx: click.Context, *args, **kwargs):
        bsm: BedrockServerManager = ctx.obj["bsm"]
        if not bsm.can_manage_services:
            os_type = bsm.get_os_type()
            if os_type == "Windows":
                msg = "Error: This command requires 'pywin32' to be installed (`pip install pywin32`) for Web UI service management."
            else:
                msg = "Error: This command requires a supported service manager (e.g., systemd for Linux), which was not found."
            click.secho(msg, fg="red")
            raise click.Abort()
        return func(*args, **kwargs)

    return wrapper


def _perform_web_service_configuration(
    bsm: BedrockServerManager,
    setup_service: Optional[bool],
    enable_autostart: Optional[bool],
):
    """
    Internal helper to apply Web UI service configurations via API calls.

    This non-interactive function is called by `configure_web_service` (when
    flags are used) and `interactive_web_service_workflow` to execute the
    actual service configuration changes. It only acts if the system has
    service management capabilities.

    Args:
        bsm (BedrockServerManager): The BedrockServerManager instance for
            capability checks.
        setup_service (Optional[bool]): If ``True``, attempts to create/update
            the Web UI system service.
        enable_autostart (Optional[bool]): If ``True``, enables autostart for the
            service; if ``False``, disables it. ``None`` means no change to
            autostart unless `setup_service` is also ``True``.

    Calls APIs:
        - :func:`~bedrock_server_manager.api.web.create_web_ui_service`
        - :func:`~bedrock_server_manager.api.web.enable_web_ui_service`
        - :func:`~bedrock_server_manager.api.web.disable_web_ui_service`

    Raises:
        click.Abort: If API calls handled by `_handle_api_response` report errors.
    """
    if not bsm.can_manage_services:
        click.secho(
            "System service manager not available. Skipping Web UI service configuration.",
            fg="yellow",
        )
        return

    if setup_service:
        # When setting up the service, enable_autostart choice (even if None)
        # is passed to create_web_ui_service, which might have its own default.
        enable_flag = (
            enable_autostart if enable_autostart is not None else False
        )  # Default to False if not specified alongside setup
        os_type = bsm.get_os_type()
        click.secho(
            f"\n--- Configuring Web UI System Service ({os_type}) ---", bold=True
        )
        response = web_api.create_web_ui_service(autostart=enable_flag)
        _handle_api_response(response, "Web UI system service configured successfully.")
    elif (
        enable_autostart is not None
    ):  # Only change autostart if setup_service is False but autostart is specified
        click.echo("Applying autostart setting to existing Web UI service...")
        if enable_autostart:
            response = web_api.enable_web_ui_service()
            _handle_api_response(response, "Web UI service enabled successfully.")
        else:
            response = web_api.disable_web_ui_service()
            _handle_api_response(response, "Web UI service disabled successfully.")


def interactive_web_service_workflow(bsm: BedrockServerManager):
    """
    Guides the user through an interactive session to configure the Web UI system service.

    Uses `questionary` to prompt for:

        - Creating/updating the system service.
        - Enabling/disabling autostart for the service.

    Args:
        bsm (BedrockServerManager): The BedrockServerManager instance.
    """
    click.secho("\n--- Interactive Web UI Service Configuration ---", bold=True)
    setup_service_choice = None
    enable_autostart_choice = None

    if bsm.can_manage_services:
        os_type = bsm.get_os_type()
        service_type_str = (
            "Systemd Service (Linux)" if os_type == "Linux" else "Windows Service"
        )
        click.secho(f"\n--- {service_type_str} for Web UI ---", bold=True)
        if os_type == "Windows":
            click.secho(
                "(Note: This requires running the command as an Administrator)",
                fg="yellow",
            )

        if questionary.confirm(
            f"Create or update the {service_type_str} for the Web UI?", default=True
        ).ask():
            setup_service_choice = True
            autostart_prompt = (
                "Enable the Web UI service to start automatically when you log in?"
                if os_type == "Linux"
                else "Enable the Web UI service to start automatically when the system boots?"
            )
            enable_autostart_choice = questionary.confirm(
                autostart_prompt, default=False
            ).ask()
    else:
        click.secho(
            "\nSystem service manager not available. Skipping Web UI service setup.",
            fg="yellow",
        )
        return

    if setup_service_choice is None:
        click.secho("No changes selected for Web UI service.", fg="cyan")
        return

    click.echo("\nApplying chosen settings for Web UI service...")
    try:
        _perform_web_service_configuration(
            bsm=bsm,
            setup_service=setup_service_choice,
            enable_autostart=enable_autostart_choice,
        )
        click.secho("\nWeb UI service configuration complete.", fg="green", bold=True)
    except BSMError as e:
        click.secho(f"Error during Web UI service configuration: {e}", fg="red")
    except (click.Abort, KeyboardInterrupt):
        click.secho("\nOperation cancelled.", fg="yellow")


@click.group()
def web():
    """
    Manages the Bedrock Server Manager Web UI application.

    This group of commands allows you to start and stop the web server,
    and to manage its integration as an OS-level system service for
    features like automatic startup.
    """
    pass


@web.command("start")
@click.option(
    "-H",
    "--host",
    "hosts",
    multiple=True,
    help="Host address to bind to. Use multiple times for multiple hosts.",
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Run in Flask's debug mode (NOT for production).",
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(["direct", "detached"], case_sensitive=False),
    default="direct",
    show_default=True,
    help="Run mode: 'direct' blocks the terminal, 'detached' runs in the background.",
)
def start_web_server(hosts: Tuple[str], debug: bool, mode: str):
    """
    Starts the Bedrock Server Manager web UI.

    This command launches the Uvicorn server that hosts the FastAPI web application.
    It can run in 'direct' mode (blocking the terminal, useful for development or
    when managed by an external process manager) or 'detached' mode (running in
    the background as a new process).

    The web server's listening host(s) and debug mode can be configured via options.

    Calls API: :func:`~bedrock_server_manager.api.web.start_web_server_api`.
    """
    click.echo(f"Attempting to start web server in '{mode}' mode...")
    if mode == "direct":
        click.secho(
            "Server will run in this terminal. Press Ctrl+C to stop.", fg="cyan"
        )

    try:
        host_list = (
            list(hosts) if hosts else None
        )  # Pass None if no hosts are provided, API handles default
        response = web_api.start_web_server_api(host_list, debug, mode)

        # In 'direct' mode, start_web_server_api (which calls bsm.start_web_ui_direct)
        # is blocking. So, we'll only reach here after it stops or if mode is 'detached'.
        if mode == "detached":
            if response.get("status") == "error":
                message = response.get("message", "An unknown error occurred.")
                click.secho(f"Error: {message}", fg="red")
                raise click.Abort()
            else:
                pid = response.get("pid", "N/A")
                message = response.get(
                    "message",
                    f"Web server start initiated in detached mode (PID: {pid}).",
                )
                click.secho(f"Success: {message}", fg="green")
        elif (
            response and response.get("status") == "error"
        ):  # Should only happen if direct mode itself fails to launch
            message = response.get(
                "message", "Failed to start web server in direct mode."
            )
            click.secho(f"Error: {message}", fg="red")
            raise click.Abort()

    except BSMError as e:  # Catch errors from API if they propagate
        click.secho(f"Failed to start web server: {e}", fg="red")
        raise click.Abort()


@web.command("stop")
def stop_web_server():
    """
    Stops a detached Bedrock Server Manager web UI process.

    This command attempts to find and terminate a web server process that was
    previously started in 'detached' mode. It typically relies on a PID file
    to identify the correct process.

    This command does not affect web servers started in 'direct' mode or those
    managed by system services.

    Calls API: :func:`~bedrock_server_manager.api.web.stop_web_server_api`.
    """
    click.echo("Attempting to stop the web server...")
    try:
        response = web_api.stop_web_server_api()
        _handle_api_response(response, "Web server stopped successfully.")
    except BSMError as e:
        click.secho(f"An error occurred: {e}", fg="red")
        raise click.Abort()


@web.group("service")
def web_service_group():
    """
    Manages OS-level service integrations for the Web UI application.

    This group contains commands to configure, enable, disable, remove, and
    check the status of the system service (systemd on Linux, Windows Service
    on Windows) for the Bedrock Server Manager Web UI.
    """
    pass


@web_service_group.command("configure")
@click.option(
    "--setup-service",
    is_flag=True,
    help="Create or update the system service file for the Web UI.",
)
@click.option(
    "--enable-autostart/--no-enable-autostart",
    "autostart_flag",
    default=None,
    help="Enable or disable Web UI service autostart.",
)
@click.pass_context
def configure_web_service(
    ctx: click.Context,
    setup_service: bool,
    autostart_flag: Optional[bool],
):
    """
    Configures the OS-level system service for the Web UI application.

    This command allows setting up the Web UI to run as a system service,
    enabling features like automatic startup on boot/login.

    If run without any specific configuration flags (`--setup-service`,
    `--enable-autostart`), it launches an interactive wizard
    (:func:`~.interactive_web_service_workflow`) to guide the user.

    If flags are provided, it applies those settings directly. The command
    respects system capabilities (e.g., won't attempt service setup if a
    service manager isn't available or `pywin32` is missing on Windows).

    Calls internal helpers:

        - :func:`~.interactive_web_service_workflow` (if no flags)
        - :func:`~._perform_web_service_configuration` (if flags are present)

    """
    bsm: BedrockServerManager = ctx.obj["bsm"]
    if setup_service and not bsm.can_manage_services:
        click.secho(
            "Error: --setup-service is not available (service manager not found).",
            fg="red",
        )
        raise click.Abort()

    try:
        no_flags_used = not setup_service and autostart_flag is None

        if no_flags_used:
            click.secho(
                "No flags provided; starting interactive Web UI service setup...",
                fg="yellow",
            )
            interactive_web_service_workflow(bsm)
            return

        click.secho("\nApplying Web UI service configuration...", bold=True)
        _perform_web_service_configuration(
            bsm=bsm,
            setup_service=setup_service,
            enable_autostart=autostart_flag,
        )
        click.secho("\nWeb UI configuration applied successfully.", fg="green")
    except MissingArgumentError as e:
        click.secho(f"Configuration Error: {e}", fg="red")
    except BSMError as e:
        click.secho(f"Operation failed: {e}", fg="red")
    except (click.Abort, KeyboardInterrupt):
        click.secho("\nOperation cancelled.", fg="yellow")


@web_service_group.command("enable")
@requires_web_service_manager
def enable_web_service_cli():
    """
    Enables the Web UI system service for automatic startup.

    Configures the OS service for the Web UI (systemd on Linux, Windows Service
    on Windows) to start automatically when the system boots or user logs in.

    Requires a supported service manager (checked by decorator).

    Calls API: :func:`~bedrock_server_manager.api.web.enable_web_ui_service`.
    """
    click.echo("Attempting to enable Web UI system service...")
    try:
        response = web_api.enable_web_ui_service()
        _handle_api_response(response, "Web UI service enabled successfully.")
    except BSMError as e:
        click.secho(f"Failed to enable Web UI service: {e}", fg="red")
        raise click.Abort()


@web_service_group.command("disable")
@requires_web_service_manager
def disable_web_service_cli():
    """
    Disables the Web UI system service from starting automatically.

    Configures the OS service for the Web UI to not start automatically.

    Requires a supported service manager (checked by decorator).

    Calls API: :func:`~bedrock_server_manager.api.web.disable_web_ui_service`.
    """
    click.echo("Attempting to disable Web UI system service...")
    try:
        response = web_api.disable_web_ui_service()
        _handle_api_response(response, "Web UI service disabled successfully.")
    except BSMError as e:
        click.secho(f"Failed to disable Web UI service: {e}", fg="red")
        raise click.Abort()


@web_service_group.command("remove")
@requires_web_service_manager
def remove_web_service_cli():
    """
    Removes the Web UI system service definition from the OS.

    .. danger::
        This is a destructive operation. The service definition will be
        deleted from the system.

    Prompts for confirmation before proceeding.
    Requires a supported service manager (checked by decorator).

    Calls API: :func:`~bedrock_server_manager.api.web.remove_web_ui_service`.
    """
    if not questionary.confirm(
        "Are you sure you want to remove the Web UI system service?", default=False
    ).ask():
        click.secho("Removal cancelled.", fg="yellow")
        return
    click.echo("Attempting to remove Web UI system service...")
    try:
        response = web_api.remove_web_ui_service()
        _handle_api_response(response, "Web UI service removed successfully.")
    except BSMError as e:
        click.secho(f"Failed to remove Web UI service: {e}", fg="red")
        raise click.Abort()


@web_service_group.command("status")
@requires_web_service_manager
def status_web_service_cli():
    """
    Checks and displays the status of the Web UI system service.

    Reports whether the service definition exists, if it's currently
    active (running), and if it's enabled for autostart.

    Requires a supported service manager (checked by decorator).

    Calls API: :func:`~bedrock_server_manager.api.web.get_web_ui_service_status`.
    """
    click.echo("Checking Web UI system service status...")
    try:
        response = web_api.get_web_ui_service_status()
        if response.get("status") == "success":
            click.secho("Web UI Service Status:", bold=True)
            click.echo(
                f"  Service Defined: {click.style(str(response.get('service_exists', False)), fg='cyan')}"
            )
            if response.get("service_exists"):
                click.echo(
                    f"  Currently Active (Running): {click.style(str(response.get('is_active', False)), fg='green' if response.get('is_active') else 'red')}"
                )
                click.echo(
                    f"  Enabled for Autostart: {click.style(str(response.get('is_enabled', False)), fg='green' if response.get('is_enabled') else 'red')}"
                )
            if response.get("message"):
                click.secho(f"  Info: {response.get('message')}", fg="yellow")
        else:
            _handle_api_response(response)
    except BSMError as e:
        click.secho(f"Failed to get Web UI service status: {e}", fg="red")
        raise click.Abort()



# bedrock_server_manager/core/server/process_mixin.py
"""Provides the :class:`.ServerProcessMixin` for the :class:`~.core.bedrock_server.BedrockServer` class.

This mixin centralizes the logic for managing the Bedrock server's underlying
system process. Its responsibilities include:

    - Starting the server process directly in the foreground (blocking call).
    - Stopping the server process, attempting graceful shutdown before force-killing.
    - Checking the current running status of the server process.
    - Sending commands to a running server (platform-specific IPC mechanisms).
    - Retrieving process resource information (CPU, memory, uptime) if ``psutil``
      is available.

It abstracts platform-specific process management details by delegating to
functions within the :mod:`~.core.system.linux` and
:mod:`~.core.system.windows` modules, as well as using utilities from
:mod:`~.core.system.process` and :mod:`~.core.system.base`.

The availability of ``psutil`` (for :meth:`.ServerProcessMixin.get_process_info`)
is indicated by the :const:`.PSUTIL_AVAILABLE` flag defined in this module.
"""
import time
from typing import Optional, Dict, Any, TYPE_CHECKING, NoReturn

# psutil is an optional dependency, but required for process management.
try:
    import psutil

    PSUTIL_AVAILABLE = True
except ImportError:
    PSUTIL_AVAILABLE = False
    psutil = None

if TYPE_CHECKING:
    # This helps type checkers understand psutil types without making it a hard dependency.
    import psutil as psutil_for_types


# Local application imports.
from ..system import linux as system_linux_proc
from ..system import windows as system_windows_proc
from ..system import process as system_process
from .base_server_mixin import BedrockServerBaseMixin
from ..system import base as system_base
from ...error import (
    ConfigurationError,
    MissingArgumentError,
    ServerNotRunningError,
    ServerStopError,
    SendCommandError,
    FileOperationError,
    ServerStartError,
    SystemError,
    BSMError,
)


class ServerProcessMixin(BedrockServerBaseMixin):
    """Provides methods for managing the Bedrock server's system process.

    This mixin extends :class:`.BedrockServerBaseMixin` and encapsulates the
    functionality related to the lifecycle and interaction with the actual
    Bedrock server executable running as a system process. It includes methods
    for starting (in foreground), stopping, checking the running state, sending
    console commands, and retrieving resource usage information.

    It achieves platform independence by delegating OS-specific operations
    to functions within the :mod:`~.core.system.linux` and
    :mod:`~.core.system.windows` modules, and uses common utilities from
    :mod:`~.core.system.process` and :mod:`~.core.system.base`.

    This mixin assumes that other mixins or the main
    :class:`~.core.bedrock_server.BedrockServer` class will provide methods like
    ``is_installed()`` (from an installation mixin) and state management methods
    like ``set_status_in_config()`` (from :class:`.ServerStateMixin`).
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the ServerProcessMixin.

        Calls ``super().__init__(*args, **kwargs)`` to participate in cooperative
        multiple inheritance. It relies on attributes (e.g., `server_name`, `logger`,
        `settings`, `server_dir`, `app_config_dir`, `os_type`) initialized by
        :class:`.BedrockServerBaseMixin`. It also implicitly depends on methods
        that may be provided by other mixins that form the complete
        :class:`~.core.bedrock_server.BedrockServer` class (e.g.,
        :meth:`~.ServerStateMixin.set_status_in_config`,
        ``is_installed`` from an installation mixin).

        Args:
            *args (Any): Variable length argument list passed to `super()`.
            **kwargs (Any): Arbitrary keyword arguments passed to `super()`.
        """
        super().__init__(*args, **kwargs)
        # Attributes like self.server_name, self.server_dir etc. are available from BedrockServerBaseMixin.
        # Methods like self.is_installed() (from InstallUpdateMixin) and
        # self.set_status_in_config() (from ServerStateMixin) are expected to be present
        # on the final composed BedrockServer object.

    def is_running(self) -> bool:
        """Checks if the Bedrock server process is currently running and verified.

        This method delegates to
        :func:`~.core.system.base.is_server_running`, which internally uses
        :func:`~.core.system.process.get_verified_bedrock_process` to check
        not only if a process with the stored PID exists, but also if that
        process matches the expected executable path and working directory for
        this server instance.

        It does not update the server's persisted status in the configuration
        but provides a real-time check of the process state.

        Returns:
            bool: ``True`` if the server process is running and verified,
            ``False`` otherwise. This includes cases where `psutil` might be
            unavailable or if checks fail due to permissions or other system errors.

        Raises:
            ConfigurationError: If the base directory for servers (`paths.servers`)
                is not configured in settings, as this is essential for path
                verification.
        """
        self.logger.debug(f"Checking if server '{self.server_name}' is running.")

        if not self.base_dir:  # From BedrockServerBaseMixin
            raise ConfigurationError(
                "Base server directory ('paths.servers') not configured, cannot check server running status."
            )

        try:
            # system_base.is_server_running itself calls core_process.get_verified_bedrock_process
            is_running_flag = system_base.is_server_running(
                self.server_name, self.server_dir, self.app_config_dir
            )
            self.logger.debug(
                f"system_base.is_server_running for '{self.server_name}' returned: {is_running_flag}"
            )
            return is_running_flag
        except (
            BSMError
        ) as e_bsm:  # Catch known BSM errors that might occur during check
            self.logger.warning(
                f"A known error occurred during is_server_running check for '{self.server_name}': {e_bsm}"
            )
            return False
        except ConfigurationError:
            # Let ConfigurationError propagate as it's critical
            raise
        except Exception as e_unexp:
            # Catch any other unexpected errors and treat as "not running" for safety
            self.logger.error(
                f"Unexpected error during is_server_running check for '{self.server_name}': {e_unexp}",
                exc_info=True,
            )
            return False

    def send_command(self, command: str) -> None:
        """Sends a command string to the running Bedrock server process.

        This method delegates to platform-specific functions for sending commands:

            - On Linux: :func:`~.core.system.linux._linux_send_command` (uses FIFO).
            - On Windows: :func:`~.core.system.windows._windows_send_command` (uses named pipes).

        Args:
            command (str): The command string to send to the server's console
                (e.g., "list", "say Hello World").

        Raises:
            MissingArgumentError: If the `command` string is empty.
            ServerNotRunningError: If :meth:`.is_running` returns ``False`` before
                attempting to send the command.
            NotImplementedError: If the current operating system is not supported
                or if the required system module (linux/windows) is not available.
            SendCommandError: For underlying failures during the send operation
                (e.g., pipe errors, `schtasks` errors).
            CommandNotFoundError: If platform-specific helper commands (`schtasks` if it was missing but passed init)
                are not found. (More likely caught during scheduler init).
            SystemError: For other unexpected system-level errors during the send.
        """
        if not command:
            raise MissingArgumentError("Command cannot be empty.")

        if not self.is_running():  # Relies on the refined is_running()
            raise ServerNotRunningError(
                f"Cannot send command: Server '{self.server_name}' is not running."
            )

        self.logger.info(
            f"Sending command '{command}' to server '{self.server_name}' on {self.os_type}..."
        )

        try:
            if self.os_type == "Linux":
                if not system_linux_proc:  # Should not happen if correctly imported
                    raise NotImplementedError(
                        "Linux system processing module not available (system_linux_proc)."
                    )
                system_linux_proc._linux_send_command(self.server_name, command)
            elif self.os_type == "Windows":
                if not system_windows_proc:  # Should not happen
                    raise NotImplementedError(
                        "Windows system processing module not available (system_windows_proc)."
                    )
                system_windows_proc._windows_send_command(self.server_name, command)
            else:
                raise NotImplementedError(
                    f"Sending commands is not supported on operating system: {self.os_type}"
                )
            self.logger.info(
                f"Command '{command}' sent successfully to server '{self.server_name}'."
            )
        except BSMError:  # Re-raise known BSM errors
            raise
        except Exception as e_unexp:  # Wrap unexpected errors
            raise SendCommandError(
                f"An unexpected error occurred while sending command to '{self.server_name}': {e_unexp}"
            ) from e_unexp

    def start(self) -> NoReturn:
        """Starts the Bedrock server process directly in the foreground (blocking).

        This method is intended for direct, interactive server execution and will
        block the calling thread until the server process terminates (either
        gracefully or due to a crash). It handles pre-start checks, updates the
        server's status in its configuration, and delegates to platform-specific
        start functions:

            - On Linux: :func:`~.core.system.linux._linux_start_server`.
            - On Windows: :func:`~.core.system.windows._windows_start_server`.

        The server's status is set to "STARTING" before launch and should be
        updated to "RUNNING" by the platform-specific function upon successful
        startup. On termination, this method ensures the status is set to
        "STOPPED" or "ERROR".

        Requires `self.is_installed()` (from another mixin) to be ``True``.

        Raises:
            ServerStartError: If the server is not installed, is already running,
                runs on an unsupported OS, or if any error occurs during the
                startup or execution of the server process. This can wrap
                various underlying exceptions like :class:`~.error.BSMError` or
                :class:`~.error.SystemError`.
        """
        # --- Pre-flight Checks ---
        if not hasattr(self, "is_installed") or not self.is_installed():
            raise ServerStartError(
                f"Cannot start server '{self.server_name}': Not installed or "
                f"invalid installation at {self.server_dir} (is_installed check failed or method missing)."
            )

        if self.is_running():
            self.logger.warning(
                f"Attempted to start server '{self.server_name}' but it is already running."
            )
            raise ServerStartError(f"Server '{self.server_name}' is already running.")

        # --- Begin Startup Process ---
        try:
            if hasattr(self, "set_status_in_config"):
                self.set_status_in_config("STARTING")  # type: ignore
            else:
                self.logger.warning(
                    "set_status_in_config method not found; cannot update status to STARTING."
                )
        except Exception as e_status:
            self.logger.warning(
                f"Failed to set status to STARTING for '{self.server_name}': {e_status}"
            )

        self.logger.info(
            f"Attempting a direct (blocking) start for server '{self.server_name}' "
            f"on {self.os_type}..."
        )

        try:
            # --- Platform-Specific Blocking Call ---
            if self.os_type == "Linux":
                system_linux_proc._linux_start_server(
                    self.server_name, self.server_dir, self.app_config_dir
                )
            elif self.os_type == "Windows":
                system_windows_proc._windows_start_server(
                    self.server_name, self.server_dir, self.app_config_dir
                )
            else:
                # This case should ideally be caught earlier if OS is unsupported for any operation.
                if hasattr(self, "set_status_in_config"):
                    self.set_status_in_config("ERROR")  # type: ignore
                raise ServerStartError(
                    f"Unsupported operating system for server start: {self.os_type}"
                )

            self.logger.info(
                f"Direct server session for '{self.server_name}' has ended gracefully."
            )

        except BSMError as e_bsm_start:  # Catch known BSM errors during startup
            self.logger.error(
                f"A known BSM error occurred while starting server '{self.server_name}': {e_bsm_start}",
                exc_info=True,
            )
            if hasattr(self, "set_status_in_config"):
                self.set_status_in_config("ERROR")  # type: ignore
            raise ServerStartError(
                f"Failed to start server '{self.server_name}': {e_bsm_start}"
            ) from e_bsm_start
        except (
            Exception
        ) as e_unexp_runtime:  # Catch unexpected errors during server runtime
            self.logger.error(
                f"An unexpected error occurred while server '{self.server_name}' was running: {e_unexp_runtime}",
                exc_info=True,
            )
            if hasattr(self, "set_status_in_config"):
                self.set_status_in_config("ERROR")  # type: ignore
            raise ServerStartError(
                f"Unexpected error during server '{self.server_name}' execution: {e_unexp_runtime}"
            ) from e_unexp_runtime
        finally:
            # --- Final Status Cleanup ---
            # Check current stored status. If it's still STARTING or RUNNING, correct it.
            if hasattr(self, "get_status_from_config") and hasattr(
                self, "set_status_in_config"
            ):
                current_stored_status = self.get_status_from_config()  # type: ignore
                if current_stored_status not in ("STOPPED", "ERROR"):
                    self.logger.info(
                        f"Server '{self.server_name}' process ended. Correcting stored status from '{current_stored_status}' to STOPPED."
                    )
                    self.set_status_in_config("STOPPED")  # type: ignore
            else:
                self.logger.warning(
                    "Status management methods (get/set_status_in_config) not found; final status may be inaccurate."
                )

    def stop(self) -> None:
        """Stops the Bedrock server process gracefully, with a forceful fallback.

        This method orchestrates the server shutdown sequence:

            1. Checks if the server is running using :meth:`.is_running`. If not,
               ensures the stored status is "STOPPED" and returns.
            2. Sets the server's stored status to "STOPPING".
            3. Attempts a graceful shutdown by sending the "stop" command via
               :meth:`.send_command`.
            4. Waits for the process to terminate, checking its status periodically
               up to a configured timeout (`SERVER_STOP_TIMEOUT_SEC` from settings).
            5. If the server is still running after the timeout, it attempts a
               forceful PID-based termination using utilities from
               :mod:`~.core.system.process`.
            6. Updates the stored status to "STOPPED" upon successful termination,
               or "ERROR" if issues persist.

        Raises:
            ServerStopError: If the server fails to stop after all attempts and
                is still detected as running.
        """
        if not self.is_running():
            self.logger.info(
                f"Attempted to stop server '{self.server_name}', but it is not currently running."
            )
            # Ensure stored status is consistent if is_running is false.
            if hasattr(self, "get_status_from_config") and hasattr(
                self, "set_status_in_config"
            ):
                if self.get_status_from_config() != "STOPPED":  # type: ignore
                    try:
                        self.set_status_in_config("STOPPED")  # type: ignore
                    except Exception as e_stat:
                        self.logger.warning(
                            f"Failed to set status to STOPPED for non-running server '{self.server_name}': {e_stat}"
                        )
            return

        try:
            if hasattr(self, "set_status_in_config"):
                self.set_status_in_config("STOPPING")  # type: ignore
        except Exception as e_stat:  # Should be more specific if possible
            self.logger.warning(
                f"Failed to set status to STOPPING for '{self.server_name}': {e_stat}"
            )

        self.logger.info(f"Attempting to stop server '{self.server_name}'...")

        # --- 1. Attempt graceful shutdown via command ---
        graceful_attempted = False
        try:
            # Ensure send_command method exists (it should if this mixin is used correctly)
            if hasattr(self, "send_command"):
                self.send_command("stop")  # This method is part of ServerProcessMixin
                self.logger.info(f"Sent 'stop' command to server '{self.server_name}'.")
                graceful_attempted = True
            else:
                self.logger.warning(  # Should not happen in normal use
                    "send_command method not found on self. Cannot send graceful stop command."
                )
        except BSMError as e_cmd:  # Catch known BSM errors from send_command
            self.logger.warning(
                f"Failed to send 'stop' command to '{self.server_name}': {e_cmd}. Proceeding to check process status."
            )
        except Exception as e_unexp_cmd:  # Catch any other unexpected error
            self.logger.error(
                f"Unexpected error sending 'stop' command to '{self.server_name}': {e_unexp_cmd}",
                exc_info=True,
            )

        # --- 2. Wait for process to terminate ---
        # Only wait if a graceful stop was attempted, or always wait if is_running was true?
        # Current logic waits regardless, which is fine.
        stop_timeout_sec = self.settings.get("SERVER_STOP_TIMEOUT_SEC", 60)
        # Make max_attempts at least 1 to ensure at least one check after command.
        max_attempts = max(1, stop_timeout_sec // 2)
        sleep_interval = 2  # seconds

        self.logger.info(
            f"Waiting up to {max_attempts * sleep_interval}s for '{self.server_name}' process to terminate..."
        )

        for attempt in range(max_attempts):
            if not self.is_running():
                if hasattr(self, "set_status_in_config"):
                    self.set_status_in_config("STOPPED")  # type: ignore
                self.logger.info(
                    f"Server '{self.server_name}' stopped successfully (detected on attempt {attempt + 1})."
                )
                return

            self.logger.debug(
                f"Waiting for '{self.server_name}' to stop (attempt {attempt + 1}/{max_attempts})..."
            )
            time.sleep(sleep_interval)

        # --- 3. If still running, attempt forceful PID-based termination ---
        if self.is_running():  # Re-check after waiting
            self.logger.error(
                f"Server '{self.server_name}' failed to stop gracefully after command and wait."
            )
            self.logger.info(
                f"Attempting forceful PID-based termination for server '{self.server_name}'."
            )

            # Ensure get_pid_file_path method exists (from BaseServerMixin or similar)
            if not hasattr(self, "get_pid_file_path"):
                self.logger.error(
                    "get_pid_file_path method not found. Cannot perform PID-based termination."
                )
                if hasattr(self, "set_status_in_config"):
                    self.set_status_in_config("ERROR")  # type: ignore
                raise ServerStopError(
                    f"Cannot perform PID-based stop for '{self.server_name}': missing PID file path method."
                )

            pid_file_path = self.get_pid_file_path()  # type: ignore
            try:
                pid_to_terminate = system_process.read_pid_from_file(pid_file_path)
                if pid_to_terminate and system_process.is_process_running(
                    pid_to_terminate
                ):
                    self.logger.info(
                        f"Terminating PID {pid_to_terminate} for '{self.server_name}'."
                    )
                    system_process.terminate_process_by_pid(pid_to_terminate)
                    time.sleep(
                        sleep_interval
                    )  # Give a moment for termination to reflect
                    if not self.is_running():  # Final check
                        if hasattr(self, "set_status_in_config"):
                            self.set_status_in_config("STOPPED")  # type: ignore
                        self.logger.info(
                            f"Server '{self.server_name}' (PID {pid_to_terminate}) forcefully terminated and confirmed stopped."
                        )
                        return
                    else:
                        self.logger.error(
                            f"Server '{self.server_name}' (PID {pid_to_terminate}) STILL RUNNING after forceful termination attempt."
                        )
                elif pid_to_terminate:
                    self.logger.info(
                        f"PID {pid_to_terminate} from file for '{self.server_name}' is not running. Removing stale PID file."
                    )
                    system_process.remove_pid_file_if_exists(pid_file_path)
                    # If it wasn't running by PID, but is_running() said it was, there's a discrepancy.
                    # However, if is_running() now returns false, we're good.
                    if not self.is_running() and hasattr(self, "set_status_in_config"):
                        self.set_status_in_config("STOPPED")  # type: ignore
                    return

            except (
                FileOperationError,
                SystemError,
                ServerStopError,
                MissingArgumentError,
            ) as e_force:
                self.logger.error(
                    f"Error during forceful termination attempt for '{self.server_name}': {e_force}",
                    exc_info=True,
                )
            except Exception as e_unexp_force:
                self.logger.error(
                    f"Unexpected error during forceful termination of '{self.server_name}': {e_unexp_force}",
                    exc_info=True,
                )
        else:  # Was not running after the wait period
            if hasattr(self, "set_status_in_config"):
                self.set_status_in_config("STOPPED")  # type: ignore
            self.logger.info(
                f"Server '{self.server_name}' confirmed stopped after waiting period."
            )
            return

        # --- 4. Final status update and error if still running ---
        if hasattr(self, "get_status_from_config") and hasattr(
            self, "set_status_in_config"
        ):
            if self.get_status_from_config() != "STOPPED":  # type: ignore
                self.set_status_in_config("ERROR")  # type: ignore

        if self.is_running():  # Final check
            raise ServerStopError(
                f"Server '{self.server_name}' failed to stop after all attempts. Manual intervention may be required."
            )
        else:
            # This path means it stopped, but possibly due to forceful kill or after initial command failed.
            # Status should have been set to STOPPED if termination was confirmed.
            # If it's ERROR here, it means forceful termination might have also failed confirmation.
            if hasattr(self, "get_status_from_config") and self.get_status_from_config() != "STOPPED":  # type: ignore
                self.logger.warning(
                    f"Server '{self.server_name}' stopped, but final status in config is not 'STOPPED'. Current: {self.get_status_from_config()}."
                )
                if hasattr(self, "set_status_in_config"):
                    self.set_status_in_config("STOPPED")  # type: ignore

    def get_process_info(self) -> Optional[Dict[str, Any]]:
        """Gets resource usage information (PID, CPU, Memory, Uptime) for the running server process.

        This method first uses
        :func:`~.core.system.process.get_verified_bedrock_process` to locate and
        verify the Bedrock server process associated with this server instance.
        If a valid process is found, it then uses the :attr:`._resource_monitor`
        (an instance of :class:`~.core.system.base.ResourceMonitor` from the base
        mixin) to calculate its current resource statistics.

        Returns:
            Optional[Dict[str, Any]]: A dictionary containing process information
            if the server is running, verified, and ``psutil`` is available.
            The dictionary has keys: "pid", "cpu_percent", "memory_mb", "uptime".
            Returns ``None`` if the server is not running, cannot be verified,
            ``psutil`` is unavailable, or if an error occurs during statistics retrieval.
            Example: ``{"pid": 1234, "cpu_percent": 15.2, "memory_mb": 256.5, "uptime": "0:10:30"}``
        """
        try:
            # 1. Find and verify the process.
            # get_verified_bedrock_process handles cases where psutil might not be available.
            process_obj: Optional["psutil_for_types.Process"] = (
                system_process.get_verified_bedrock_process(
                    self.server_name, self.server_dir, self.app_config_dir
                )
            )

            if process_obj is None:
                self.logger.debug(
                    f"No verified process found for server '{self.server_name}' to get info."
                )
                return None

            # 2. Delegate the measurement of the found process to the resource monitor.
            # _resource_monitor is initialized in BedrockServerBaseMixin.
            # It also checks for PSUTIL_AVAILABLE.
            return self._resource_monitor.get_stats(process_obj)

        except (
            BSMError
        ) as e_bsm:  # Catch known BSM errors, e.g. from get_verified_bedrock_process
            self.logger.warning(
                f"Known error while trying to get process info for '{self.server_name}': {e_bsm}"
            )
            return None
        except Exception as e_unexp:  # Catch any other unexpected errors
            self.logger.error(
                f"Unexpected error getting process info for '{self.server_name}': {e_unexp}",
                exc_info=True,
            )
            return None
